/*
 *	Armadillo-2x0 GPIO Driver
 *	Copyright (C) 2005-2006,2009 Atmark Techno, Inc.
 */

#include <linux/autoconf.h>
#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/irq.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <asm/io.h>

#include <asm/arch/ep93xx-regs.h>
#include <asm/arch/platform.h>
#include <asm/arch/armadillo2x0_gpio.h>

#define AUTHOR      "Atmark Techno, Inc."
#define DESCRIPTION "Armadillo-2x0 GPIO driver"
#define LICENSE     "GPL v2"

#define GPIO_MINOR   (185)
#define GPIO_DEVICE  "gpio"

#define VERSION     GPIO_DEVICE ": " DESCRIPTION ", (C) 2005-2006,2009 " AUTHOR

#undef DEBUG
#if defined(DEBUG)
#define KDEBUG(args...) printk("gpio: " args)
#else
#define KDEBUG(args...)
#endif

struct wait_handler{
	wait_queue_head_t queue;
	unsigned long wait_list;
	unsigned long int_list;

	struct wait_handler *prev;
	struct wait_handler *next;
};

static struct wait_handler *wait_list_head;

struct gpio_port_reg {
	unsigned long ddr;
	unsigned long dr;
	unsigned long inten;
	unsigned long inttype1;
	unsigned long inttype2;
	unsigned long eoi;
	unsigned long db;
	unsigned long rawintsts;
	unsigned long intsts;
};

static struct gpio_port_reg porta = {
	.ddr       = EP93XX_GPIO_A_DIRECTION,
	.dr        = EP93XX_GPIO_A_DATA,
	.inten     = EP93XX_GPIO_A_INT_ENABLE,
	.inttype1  = EP93XX_GPIO_A_INT_TYPE1,
	.inttype2  = EP93XX_GPIO_A_INT_TYPE2,
	.eoi       = EP93XX_GPIO_A_INT_ACK,
	.db        = EP93XX_GPIO_A_DEBOUNCE,
	.rawintsts = EP93XX_GPIO_A_RAW_INT_STATUS,
	.intsts    = EP93XX_GPIO_A_INT_STATUS,
};

static struct gpio_port_reg portb = {
	.ddr       = EP93XX_GPIO_B_DIRECTION,
	.dr        = EP93XX_GPIO_B_DATA,
	.inten     = EP93XX_GPIO_B_INT_ENABLE,
	.inttype1  = EP93XX_GPIO_B_INT_TYPE1,
	.inttype2  = EP93XX_GPIO_B_INT_TYPE2,
	.eoi       = EP93XX_GPIO_B_INT_ACK,
	.db        = EP93XX_GPIO_B_DEBOUNCE,
	.rawintsts = EP93XX_GPIO_B_RAW_INT_STATUS,
	.intsts    = EP93XX_GPIO_B_INT_STATUS,
};

static struct gpio_port_reg portc = {
	.ddr       = EP93XX_GPIO_C_DIRECTION,
	.dr        = EP93XX_GPIO_C_DATA,
};

static struct gpio_port_reg portd = {
	.ddr       = EP93XX_GPIO_D_DIRECTION,
	.dr        = EP93XX_GPIO_D_DATA,
};

struct gpio_info {
	unsigned long no;
	struct gpio_port_reg *port;
	unsigned long bit;
};

static struct gpio_info gpio[] = {
	{GPIO0, &porta, BIT(4)},
	{GPIO1, &porta, BIT(5)},
	{GPIO2, &porta, BIT(6)},
	{GPIO3, &porta, BIT(7)},
	{GPIO4, &portb, BIT(0)},
	{GPIO5, &portb, BIT(1)},
	{GPIO6, &portb, BIT(2)},
	{GPIO7, &portb, BIT(3)},
	{GPIO8, &portd, BIT(0)},
	{GPIO9, &portd, BIT(1)},
	{GPIO10,&portd, BIT(2)},
	{GPIO11,&portd, BIT(3)},
	{GPIO12,&portc, BIT(0)},
	{GPIO13,&portc, BIT(1)},
	{GPIO14,&portc, BIT(2)},
	{GPIO15,&portc, BIT(3)},
};

static int gpio_num;

#define change_reg_bit(REG,BIT,VALUE)       \
({                                          \
	unsigned tmp;                       \
	tmp = (inl(REG) & ~BIT);            \
	outl(tmp | (VALUE ? BIT : 0), REG); \
})

#define get_reg_bit(REG,BIT)                \
	inl(REG) & BIT

#if defined(DEBUG)
static void debug_wait_handle(struct wait_handler *head)
{
	int count = 0;
	struct wait_handler *tmp;

	KDEBUG("debug_wait_handle");
	KDEBUG("display wait_handle_list\n");
	if (!head){
		printk("list not found\n");
		return;
	}
	
	for (tmp = head; tmp != NULL; tmp = tmp->next) {
		printk("chain[%2d]: 0x%08lx\n", count++, (unsigned long)tmp);
		printk("      prev: 0x%08lx\n", (unsigned long)tmp->prev);
		printk("      next: 0x%08lx\n", (unsigned long)tmp->next);
		printk("\n");
	}
}
#else
#define debug_wait_handle(head)
#endif

static struct wait_handler *create_wait_handle(unsigned long wait_list)
{
	struct wait_handler *tmp;
	struct wait_handler *handle;

	KDEBUG("create_wait_handle\n");
	handle = (struct wait_handler *)kmalloc(sizeof(struct wait_handler), 
						GFP_KERNEL);
	if (!handle)
		return NULL;
	
	init_waitqueue_head(&handle->queue);
	handle->wait_list = wait_list;
	handle->int_list = 0;
	handle->next = NULL;
	
	if (wait_list_head) {
		tmp = wait_list_head;    

		while(tmp->next)
			tmp = tmp->next;

		tmp->next = handle;
		handle->prev= tmp;
	} else {
		wait_list_head = handle;
		handle->prev = NULL;
	}
	
	debug_wait_handle(wait_list_head);

	return handle;
}

static void destroy_wait_handle(struct wait_handler **handle)
{
	struct wait_handler *tmp;

	KDEBUG("destroy_wait_handle\n");
	tmp = wait_list_head;
	
	if(!*handle)
		return;
	
	while (tmp != *handle) {
		tmp = tmp->next;
		if(!tmp)
			return;
	}
	
	if(tmp->prev)
		tmp->prev->next = tmp->next;

	if(tmp->next)
		tmp->next->prev = tmp->prev;

	if(!tmp->prev)
		wait_list_head = tmp->next;

	kfree(*handle);
	
	*handle = NULL;
	
	debug_wait_handle(wait_list_head);
}

static void gpio_irq_handler(unsigned int irq, struct irq_desc *desc)
{
	struct wait_handler *handle;
	unsigned long int_sts_a;
	unsigned long int_sts_b;
	int i;
	
	KDEBUG("gpio_irq_handler\n");
	
	int_sts_a = inl(porta.intsts);
	int_sts_b = inl(portb.intsts);
	
	if (!wait_list_head)
		goto no_wait_list;

	for (handle = wait_list_head; handle != NULL; handle = handle->next) {
		for (i = 0; i < gpio_num; i++) {
			if (handle->wait_list & gpio[i].no && gpio[i].port->inten){
				if (get_reg_bit(gpio[i].port->intsts, gpio[i].bit)) {
					handle->int_list = gpio[i].no;
					wake_up_interruptible(&handle->queue);
				}
			}
		}
	}
	
 no_wait_list:
	outl(int_sts_a, porta.eoi);
	outl(int_sts_b, portb.eoi);
}

static int gpio_open(struct inode *inode, struct file *file)
{
	KDEBUG("gpio_open\n");

	return 0;
}

static int gpio_release(struct inode *inode, struct file *file)
{
	KDEBUG("gpio_release\n");

	return 0;
}

static unsigned long get_gpio_num(unsigned long no)
{
	int i;

	for (i = 0; i < gpio_num; i++)
		if (no == gpio[i].no)
			break;

	return i;
}

static int set_param(struct gpio_param *param)
{
	unsigned long num;

	KDEBUG("set_param\n");
	
	switch (param->mode) {
	case MODE_OUTPUT:
		KDEBUG("MODE_OUTPUT\n");

		num = get_gpio_num(param->no);
		if (num == gpio_num)
			return -1;

		//DIRECTION <= OUTPUT
		change_reg_bit(gpio[num].port->ddr, gpio[num].bit, 1);

		//DATA <= value
		change_reg_bit(gpio[num].port->dr, gpio[num].bit,
			       param->data.o.value);
				
		KDEBUG("ditect GPIO%d (%s)\n", num,
		       (param->data.o.value ? "ON" : "OFF"));
		break;

	case MODE_INPUT:
		KDEBUG("MODE_INPUT\n");

		num = get_gpio_num(param->no);
		if (num == gpio_num)
			return -1;

		//DDR <= INPUT
		change_reg_bit(gpio[num].port->ddr, gpio[num].bit, 0);

		if (param->data.i.int_enable && gpio[num].port->inten) {
			set_irq_flags(IRQ_EP93XX_GPIO_AB, 0);
			//INTTYPE1,INTTYPE2 <= interrupt type
			switch (param->data.i.int_type & 0x03) {
			case TYPE_LOW_LEVEL:
				change_reg_bit(gpio[num].port->inttype1,
					       gpio[num].bit, 0);
				change_reg_bit(gpio[num].port->inttype2,
					       gpio[num].bit, 0);
				break;
			case TYPE_HIGH_LEVEL:
				change_reg_bit(gpio[num].port->inttype1,
					       gpio[num].bit, 0);
				change_reg_bit(gpio[num].port->inttype2,
					       gpio[num].bit, 1);
				break;
			case TYPE_FALLING_EDGE:
				change_reg_bit(gpio[num].port->inttype1,
					       gpio[num].bit, 1);
				change_reg_bit(gpio[num].port->inttype2,
					       gpio[num].bit, 0);
				break;
			case TYPE_RISING_EDGE:
				change_reg_bit(gpio[num].port->inttype1,
					       gpio[num].bit, 1);
				change_reg_bit(gpio[num].port->inttype2,
					       gpio[num].bit, 1);
				break;
			}

			//DB <= debounce
			if (param->data.i.int_type & TYPE_DEBOUNCE) {
				change_reg_bit(gpio[num].port->db,
					       gpio[num].bit, 1);
				msleep(10);
			}

			//EOI <= clear
			change_reg_bit(gpio[num].port->eoi,
				       gpio[num].bit, 1);
					
			//INTEN <= enable
			change_reg_bit(gpio[num].port->inten,
				       gpio[num].bit, 1);

			set_irq_flags(IRQ_EP93XX_GPIO_AB, IRQF_VALID | IRQF_PROBE);
		} else {
			if (gpio[num].port->eoi)
				change_reg_bit(gpio[num].port->eoi,
					       gpio[num].bit, 1);

			if (gpio[num].port->inten)
				change_reg_bit(gpio[num].port->inten,
					       gpio[num].bit, 0);

			if (gpio[num].port->inttype1)
				change_reg_bit(gpio[num].port->inttype1,
					       gpio[num].bit, 0);

			if (gpio[num].port->inttype2)
				change_reg_bit(gpio[num].port->inttype2,
					       gpio[num].bit, 0);

			if (gpio[num].port->db)
				change_reg_bit(gpio[num].port->db,
					       gpio[num].bit, 0);
		}
				
		KDEBUG("ditect GPIO%d\n", num);
		break;

	default:
		return -1;
	}

	return 0;
}

static int get_param(struct gpio_param *param)
{
	unsigned long num;
	unsigned long io;
	
	KDEBUG("get_param\n");
	
	num = get_gpio_num(param->no);
	if (num == gpio_num)
		return -EINVAL;

	io = get_reg_bit(gpio[num].port->ddr, gpio[num].bit);
	if (io) {
		//MODE_OUTPUT => mode
		param->mode = MODE_OUTPUT;

		//DR => value
		param->data.o.value =
			(get_reg_bit(gpio[num].port->dr,
				     gpio[num].bit) ? 1 : 0); 
				
		KDEBUG("ditect GPIO%d (%s)\n", num,
		       (param->data.o.value ? "ON" : "OFF"));
	} else {
		//MODE_INPUT => mode
		param->mode = MODE_INPUT;

		//DR => value
		param->data.i.value =(get_reg_bit(gpio[num].port->dr,
						  gpio[num].bit) ? 1 : 0);
				
		if (gpio[num].port->inten) {
			//INTEN => inten
			param->data.i.int_enable =
				(get_reg_bit(gpio[num].port->inten,
					     gpio[num].bit) ? 1 : 0);
				  
			if(param->data.i.int_enable){
				//INTTYPE1,INTTYPE2 => interrupt type
				param->data.i.int_type =
					(get_reg_bit(gpio[num].port->inttype1,
						     gpio[num].bit) ? BIT(1) : 0);
				param->data.i.int_type |=
					(get_reg_bit(gpio[num].port->inttype2,
						     gpio[num].bit) ? BIT(0) : 0);
					
				//DB => debounce
				param->data.i.int_type |=
					(get_reg_bit(gpio[num].port->db,
						     gpio[num].bit) ? BIT(2) : 0);
			} else
				param->data.i.int_type = 0;
		}
				
		KDEBUG("ditect GPIO%d\n", num);
	}

	return 0;
}

static int gpio_ioctl(struct inode *inode, struct file *file,
		      unsigned int cmd,unsigned long arg)
{
	struct wait_handler *wait_handle;
	struct wait_param *wait_param;
	struct gpio_param *param_list;
	struct gpio_param *param;
	int ret;
	
	KDEBUG("gpio_ioctl\n");
	
	switch (cmd) {
	case PARAM_SET:
		KDEBUG("PARAM_SET\n");
		param_list = (struct gpio_param *)arg;
		
		for (param = param_list; param != NULL; param = param->next) {
			ret = set_param(param);
			if (ret)
				return -EINVAL;
		}

		break;
	case PARAM_GET:
		KDEBUG("PARAM_GET\n");
		param_list = (struct gpio_param *)arg;
		
		
		for (param = param_list; param != NULL; param = param->next) {
			ret = get_param(param);
			if (ret)
				return -EINVAL;
		}

		break;
	case INTERRUPT_WAIT:
		wait_param = (struct wait_param *)arg;
		KDEBUG("INTERRUPT_WAIT\n");
		
		wait_handle = create_wait_handle(wait_param->list);
		if (!wait_handle)
			return -EBUSY;
		
		if (wait_param->timeout) {
			interruptible_sleep_on_timeout(&wait_handle->queue,
						       //msec->jiffies
						       (wait_param->timeout * HZ + 999)
						       / 1000);
		} else
			interruptible_sleep_on(&wait_handle->queue);
		
		wait_param->list = wait_handle->int_list;
		
		destroy_wait_handle(&wait_handle);
		
		break;
	default:
		return -EINVAL;
	}
	
	return 0;
}

static struct file_operations gpio_fops = {
	.owner   = THIS_MODULE,
	.open    = gpio_open,
	.release = gpio_release,
	.ioctl   = gpio_ioctl,
};

static struct miscdevice gpio_dev = {
	.minor = GPIO_MINOR,
	.name  = GPIO_DEVICE,
	.fops  = &gpio_fops,
};

static int __init gpio_init(void)
{
	KDEBUG("gpio_init\n");

	switch (get_board_id()) {
	case BOARD_ID_A210:
	case BOARD_ID_A210C:
		gpio_num = GPIO_NUM_A210;
		break;
	case BOARD_ID_A220:
	case BOARD_ID_A230:
	case BOARD_ID_A240:
		gpio_num = GPIO_NUM;
		break;
	}

	if (misc_register(&gpio_dev))
		return -ENODEV;

	printk(KERN_INFO VERSION "\n");

	set_irq_handler(IRQ_EP93XX_GPIO_AB, gpio_irq_handler);

	return 0;
}

static void __exit gpio_exit(void)
{
	KDEBUG("gpio_exit\n");
	set_irq_handler(IRQ_EP93XX_GPIO_AB, NULL);
	misc_deregister(&gpio_dev);
}

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESCRIPTION);
MODULE_LICENSE(LICENSE);

module_init(gpio_init);
module_exit(gpio_exit);
